/*
 * rshaper.c
 *
 * Copyright (c) 1998-2001  Alessandro Rubini <rubini@linux.it>
 * Copyright (c) 2001       Rodolfo Giometti <giometti@linux.it>
 * Copyright (c) 2008       Vladimir Oleynik <dzo@simtreas.ru>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */

// #define RSHAPER_DEBUG

#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE
#  define MODULE
#endif

//#include <linux/config.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/kernel.h> /* printk() */

#include <linux/slab.h> /* kmalloc() */
#include <linux/errno.h>  /* error codes */
#include <linux/types.h>  /* size_t */
#include <linux/interrupt.h> /* mark_bh */

#include <linux/netdevice.h>   /* struct device, and other headers */
#include <linux/etherdevice.h> /* eth_type_trans */
#include <linux/in.h>
#include <linux/ip.h>          /* struct iphdr */
#include <linux/tcp.h>         /* struct tcphdr */
#include <linux/skbuff.h>
#include <linux/capability.h>


#include "sysdep.h"
#include "rshaper.h"

#define NF_IP_PRE_ROUTING     0
#define NF_IP_POST_ROUTING    4

#undef PDEBUG             /* undef it, just in case */
#ifdef RSHAPER_DEBUG
#  ifdef __KERNEL__
     /* This one if debugging is on, and kernel space */
#    define PDEBUG(fmt, args...) printk( KERN_DEBUG "rshaper: " fmt, ## args)
#  else
     /* This one for user space */
#    define PDEBUG(fmt, args...) fprintf(stderr, fmt, ## args)
#  endif
#else
#  define PDEBUG(fmt, args...) /* not debugging: nothing */
#endif


static int mode = 1; /* Incoming */

/* Load time parameter */
module_param(mode, int, 0);
MODULE_PARM_DESC(mode, "Mode");

/* ------------------------------------------------ Data Structures */

/* Each queued packet */
struct rshaper_packet {
    struct sk_buff *skb;
    int (*okfn)(struct sk_buff *);              /* reinject function */
    unsigned long timestamp;
    struct rshaper_packet *next;
};

/* Each queue */
struct rshaper_queue {
    struct rshaper_packet *head;        /* first packet, to dequeue */
    struct rshaper_packet *tail;        /* last packet, to enqueue */
    struct timer_list timer;            /* use this to dequeue */
    unsigned long nexttick;             /* foreseen next tick */
    int approx;                         /* leftover bytes */
    spinlock_t lock;                    /* mutex lock */
};

/* Each host or host group */
struct rshaper_host {
    u32 ip;
    u32 mask;
    u16 portmin;
    u16 portmax;
    struct rshaper_queue *tx_queue;
    struct rshaper_queue *rx_queue;
    struct rshaper_host *next;
    int bytes_per_tick;                         /* the bandwidth in B/tick */
    int timeout; /* measured in jiffies */
};

static struct rshaper_host *first_host;
static spinlock_t rshaper_host_lock;


/* ------------------------------------------------ Queue functions */

/* this is entered with the queue lock held */
static int enqueue_packet(struct sk_buff *skb,
			      int (*okfn)(struct sk_buff *),
			      struct rshaper_queue *q,
			      int bytes_per_tick)
{
    struct rshaper_packet *pkt;
    int size = skb_tail_pointer(skb) - skb->data;
    int timerflag=0;

    /* if the `next tick time' is expired, transmit it now */
    if (time_before_eq(q->nexttick, jiffies)) {
	PDEBUG("enqueue: now\n");
	if (time_before(q->nexttick, jiffies))
	    q->approx = 0;
	size += q->approx;
	q->nexttick = jiffies + size / bytes_per_tick;
	q->approx = size % bytes_per_tick;

	return NF_ACCEPT; /* for 2.4 */
    }

    /* allocate a structure */
    pkt = kmalloc(sizeof(struct rshaper_packet), GFP_ATOMIC);
    if (!pkt) {
	printk(KERN_INFO "rshaper: allocation failed\n");
	kfree_skb(skb);
	return NF_STOLEN; /* ENOMEM */
    }

    /* fill it */
    pkt->skb = skb;
    pkt->okfn = okfn;
    pkt->next = NULL;
    pkt->timestamp = q->nexttick;

    PDEBUG("enqueue: %04i+%03i bytes, %3i tick, %3i left (delay %li j)\n",
	   size, q->approx, (size+q->approx)/bytes_per_tick,
	   (size+q->approx)%bytes_per_tick,
	   q->nexttick - jiffies);
    size += q->approx;
    q->nexttick += size / bytes_per_tick;
    q->approx = size % bytes_per_tick;

    if (q->head) {
	PDEBUG("updating tail (h %p, t %p)\n",q->head,q->tail);
	q->tail->next = pkt;
	q->tail = pkt;
    } else {
	PDEBUG("registering the timer to run at %li\n",pkt->timestamp);
	q->head = q->tail = pkt;
	q->timer.expires = pkt->timestamp;
	timerflag++;
    }

    if (timerflag)
	add_timer(&q->timer);
    return NF_STOLEN;
}

/* This is called by the timer, so we must get the lock ourselves */
static void dequeue_packet(unsigned long data)
{
    struct rshaper_queue *q = (struct rshaper_queue *)data;
    struct rshaper_packet *p;
    unsigned long flags;

    PDEBUG("timer running: now %li\n",jiffies);

    spin_lock_irqsave(&q->lock, flags);

    /* consume expired packets */
    while ((p = q->head) && time_before_eq(p->timestamp, jiffies)) {
	q->head = p->next;
	if (q->tail == p)
	    q->tail = NULL;

	/* release the lock during reinjection, then reacquire it */
	spin_unlock_irqrestore(&q->lock, flags);

	p->okfn(p->skb);
	kfree(p);

	spin_lock_irqsave(&q->lock, flags);
    }

    /* if some more is there, schedule it */
    if (p) {
	q->timer.expires = p->timestamp;
	add_timer(&q->timer); /* won't be run before we return */
    }
    spin_unlock_irqrestore(&q->lock, flags);
}


/* ------------------------------------------------ Limit the queue */

#ifndef RSHAPER_LIMIT_KB
#  define RSHAPER_LIMIT_KB 64 /* half a meg per queue */
#endif

/* return a new timeout (a jiffy value) according to the max queue size */
static int rshaper_retimeout(int timeout, int bytes_per_tick)
{
    int maxticks = RSHAPER_LIMIT_KB * 1024 / bytes_per_tick;
    if (maxticks < timeout) timeout = maxticks;
    return timeout;
}

/* ------------------------------------------------ Host Management */

/* this is called with the host lock held */
static struct rshaper_host *rshaper_find_host (u32 ip, u16 portmin, u16 portmax, struct rshaper_host ***prevp)
{
    struct rshaper_host *host, **prev = &first_host;

    PDEBUG("Finding host %d.%d.%d.%d:%d-%d\n", NIPQUAD(ip), portmin, portmax);
    while (*prev && (((*prev)->ip != ip) ||
		     (*prev)->portmin != portmin || (*prev)->portmax != portmax))
	prev = &((*prev)->next);

    if (!*prev) {
	*prevp = NULL;
	return NULL;
    }
    host = *prev;
    *prevp = prev;
    return host;
}

/* this is called with the host lock held */
static int rshaper_add_host(u32 ip, u32 mask, u16 portmin, u16 portmax, int bytes_per_tick, int timeout)
{
    struct rshaper_host *host;

    host = kmalloc(sizeof(struct rshaper_host), GFP_ATOMIC);
    if (!host) return -ENOMEM;

    PDEBUG("Adding host %d.%d.%d.%d:%d-%d, %i bpt, %i jiffies\n",
	   NIPQUAD(ip), portmin, portmax, bytes_per_tick, timeout);

    host->ip = ip;
    host->mask = mask;
    host->portmin = portmin;
    host->portmax = portmax;
    host->bytes_per_tick = bytes_per_tick;
    host->timeout = timeout;
    host->tx_queue = kmalloc(sizeof(struct rshaper_queue), GFP_ATOMIC);
    host->rx_queue = kmalloc(sizeof(struct rshaper_queue), GFP_ATOMIC);

    if (!host->tx_queue || !host->rx_queue) {
	if (host->tx_queue) kfree(host->tx_queue);
	if (host->rx_queue) kfree(host->rx_queue);
	kfree(host);
	return -ENOMEM;
    }
    memset(host->rx_queue, 0, sizeof(struct rshaper_queue));
    init_timer(&host->rx_queue->timer);
    host->rx_queue->timer.data = (unsigned long)host->rx_queue;
    host->rx_queue->timer.function = dequeue_packet;
    spin_lock_init(&host->rx_queue->lock);

    memset(host->tx_queue, 0, sizeof(struct rshaper_queue));
    init_timer(&host->tx_queue->timer);
    host->tx_queue->timer.data = (unsigned long)host->tx_queue;
    host->tx_queue->timer.function = dequeue_packet;
    spin_lock_init(&host->tx_queue->lock);

    host->next = first_host;
    first_host = host;
    PDEBUG("Registered with tx %p, rx %p\n", host->tx_queue, host->rx_queue);
    return 0;

}

/* this is called after removing the host from the list (host lock held) */
static void rshaper_kfree_host(struct rshaper_host *host)
{
    struct rshaper_packet *packet, *nextp;

    del_timer_sync(&host->tx_queue->timer);
    del_timer_sync(&host->rx_queue->timer);

    /* no need to held the queue lock, as nobody will add/remove packets */
    for (packet = host->tx_queue->head; packet; ) {
	nextp = packet->next;
	kfree(packet);
	packet = nextp;
    }

    kfree(host->tx_queue);

    for (packet = host->rx_queue->head; packet; ) {
	nextp = packet->next;
	kfree(packet);
	packet = nextp;
    }

    kfree(host->rx_queue);
    PDEBUG("free host %p\n", host);
    kfree(host);
}

static inline int rshaper_chk_addr_port(u32 addr, u16 port, struct rshaper_host *ptr)
{
    /* check for addr and port */
    if (ptr->portmax && ptr->mask)
	return (ptr->portmin <= port) && (ptr->portmax >= port) &&
	       (ptr->ip == (addr & ptr->mask));

    /* check for port only */
    if (ptr->portmax)
	return (ptr->portmin <= port) && (ptr->portmax >= port);

    /* check for addr only */
    if (ptr->mask)
	return ptr->ip == (addr & ptr->mask);

    return 0;
}


/* ------------------------------------------------ Main Hook */

static unsigned int rshaper_rx_hook(unsigned int hooknum,
			     struct sk_buff *skb,
			     const struct net_device *in,
			     const struct net_device *out,
			     int (*okfn)(struct sk_buff *))
{
    /* no need to skip eth header, as this is after eth_type_trans() */
    struct iphdr *ih = (struct iphdr *)skb_network_header(skb);
    struct rshaper_host *ptr;
    u32 saddr = ih->saddr;
    u32 daddr = ih->daddr;
    u16 sport = 0;
    u16 dport = 0;
    unsigned long flags;
    int retval = NF_ACCEPT;


    /* try to get the port numbers */
    if (ih->protocol == IPPROTO_TCP
	|| ih->protocol == IPPROTO_UDP) {
	struct tcphdr *tcp=(struct tcphdr *)((__u32 *)ih+ih->ihl);
	sport = ntohs(tcp->source);
	dport = ntohs(tcp->dest);

	PDEBUG("packet: from %d.%d.%d.%d:%d to %d.%d.%d.%d:%d\n",
	       NIPQUAD(saddr), sport, NIPQUAD(daddr), dport);
    }
    else
	PDEBUG("packet: from %d.%d.%d.%d to %d.%d.%d.%d\n",
	       NIPQUAD(saddr), NIPQUAD(daddr));

    /*
     * Ok, now we must check out our list of managed hosts, and delay
     * reception if found
     */
    for (ptr = first_host; ptr; ptr = ptr->next) {
	if (rshaper_chk_addr_port(saddr, sport, ptr)) {
	    PDEBUG("Enqueue tx: %d.%d.%d.%d:%d\n", NIPQUAD(saddr), sport);
	    if (time_before_eq(ptr->tx_queue->nexttick,
		    (jiffies + ptr->timeout))) {
		spin_lock_irqsave(&ptr->tx_queue->lock, flags);
		retval =
		  enqueue_packet(skb, okfn, ptr->tx_queue, ptr->bytes_per_tick);
		spin_unlock_irqrestore(&ptr->tx_queue->lock, flags);
	    } else {
		PDEBUG("Tx queue too long (%li+%i < %li): dropped\n",
		       jiffies, ptr->timeout, ptr->tx_queue->nexttick);

		retval = NF_DROP;
	    }
	    return retval;
	}
	if (rshaper_chk_addr_port(daddr, dport, ptr))  {
	    PDEBUG("Enqueue rx: %d.%d.%d.%d:%d\n", NIPQUAD(daddr), dport);
	    if (time_before_eq(ptr->rx_queue->nexttick,
		    (jiffies + ptr->timeout))) {
			spin_lock_irqsave(&ptr->rx_queue->lock, flags);
		retval =
		  enqueue_packet(skb, okfn, ptr->rx_queue, ptr->bytes_per_tick);
			spin_unlock_irqrestore(&ptr->rx_queue->lock, flags);

	    } else {
		PDEBUG("Rx queue too long (%li+%i < %li): dropped\n",
		       jiffies, ptr->timeout, ptr->rx_queue->nexttick);

		retval = NF_DROP;
	    }
	    return retval;
	}
    }

    /* no delay, receive it right now */
    return retval;
}

static struct nf_hook_ops rshaper_rx_hook_ops = {   /* The pre-routing hook */
   .hook =       rshaper_rx_hook,
   .pf   =       PF_INET,
   .hooknum =    NF_IP_PRE_ROUTING,
   .owner  = THIS_MODULE,
};

static struct nf_hook_ops rshaper_tx_hook_ops = {   /* The post-routing hook */
   .hook =      rshaper_rx_hook,   /* Yes! It's the same as above! ;-p */
   .pf =        PF_INET,
   .hooknum =   NF_IP_POST_ROUTING,
   .owner   = THIS_MODULE,
};

/* ------------------------------------------------ User Cfg */

/*
 * There are several ways to talk to user space, devices, network interfaces,
 * /proc/entry points... Here I chose to use ioctl() on a network interface.
 *
 * SIOCDEVPRIVATE on rshaper is used to set host information,
 * SIOCDEVPRIVATE+1 is used to retrieve information.
 * SIOCDEVPRIVATE+2 is used to erase all information.
 */
static int rshaper_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd)
{
    struct rshaper_userinfo info,
	 *uptr = (struct rshaper_userinfo *)ifr->ifr_data;
    struct rshaper_host *host, **prev;
    unsigned long flags;
    int i, bw;

    if (cmd != SIOCRSHAPERERASEALLINFO) {
    if (!access_ok(VERIFY_READ, uptr, sizeof(*uptr)))
	return -EFAULT;

	if (copy_from_user(&info, uptr, sizeof(info)))
		return -EFAULT;
    }

    switch(cmd) {
      case SIOCRSHAPERSETINFO: /* set or clear one host */
	    if (!capable(CAP_NET_ADMIN))
		    return -EPERM;

	PDEBUG("ioctl: host %d.%d.%d.%d:%d-%d, count %i, to %i ms\n",
	       NIPQUAD(info.ip), info.portmin, info.portmax,
	       (int)info.bytes_per_second, info.timeout);

	/* Some sanity checks */
	    if (info.bytes_per_second && info.bytes_per_second<HZ)
		return -EINVAL;
	bw = (info.bytes_per_second+HZ/2)/HZ;

	    if (info.timeout < 0 || info.timeout > 30000)
		return -EINVAL;
	info.timeout = info.timeout * HZ / 1000; /* ms to jiffies */

	spin_lock_irqsave(&rshaper_host_lock, flags);
	    host = rshaper_find_host((u32)(info.ip),
		    (u16)(info.portmin), (u16)(info.portmax), &prev);

	if (host) {
	    if (info.bytes_per_second)  {
		/* turn byterate into bytes/tick, use nearest value */
		host->bytes_per_tick = bw;
		host->timeout = /* have a limit on the queue length */
		    rshaper_retimeout(info.timeout, bw);
	    } else { /* remove host */
		*prev = host->next;
		rshaper_kfree_host(host);
	    }
	}
	else if (info.bytes_per_second) {
	    rshaper_add_host(info.ip, info.mask, info.portmin, info.portmax,
			     bw, rshaper_retimeout(info.timeout, bw));
	}

	spin_unlock_irqrestore(&rshaper_host_lock, flags);

	    break;

      case SIOCRSHAPERGETINFO: /* return information about the host */
	if (!access_ok(VERIFY_WRITE, uptr, sizeof(*uptr)))
	    return -EFAULT;

	spin_lock_irqsave(&rshaper_host_lock, flags);
	for (host = first_host, i=0; host && i < info.ip;
	     host = host->next, i++)
	    /* nothing */;
	if (host) {
	    info.ip = (__u32)(host->ip);
	    info.mask = (__u32)(host->mask);
	    info.portmin = (__u16)(host->portmin);
	    info.portmax = (__u16)(host->portmax);
	    info.bytes_per_second = HZ * host->bytes_per_tick;
	    PDEBUG("to %i\n", host->timeout);
	    info.timeout = host->timeout * 1000/HZ; /* milliseconds */
	}

	spin_unlock_irqrestore(&rshaper_host_lock, flags);

	    if (!host)
		return -ENODEV;

	PDEBUG("ioctl-ret: host %d.%d.%d.%d:%d-%d (%08lx), count %i, to %i\n",
	       NIPQUAD(info.ip), info.portmin, info.portmax, (long)info.mask,
	       info.bytes_per_second, info.timeout);

	    if (copy_to_user(uptr, &info, sizeof(info)))
		return -EFAULT;

	    break;

	case SIOCRSHAPERERASEALLINFO: /* erase all shaping rules */
	    if (!capable(CAP_NET_ADMIN)) return -EPERM;

	    PDEBUG("ioctl: erasing all\n");
	    spin_lock_irqsave(&rshaper_host_lock, flags);
	    while ( (host = first_host) ) { /* free it all */
		    first_host = host->next;
		    rshaper_kfree_host(host);
	    }

	    spin_unlock_irqrestore(&rshaper_host_lock, flags);
	    break;

      default:
	return -ENOTTY;
    }

    return 0;
}

/* ------------------------------------------------ The Network Device */

static void rshaper_setup(struct net_device *dev)
{
    // nothing more than ioctl
    dev->do_ioctl = rshaper_ioctl;
}

static struct net_device *rshaper_dev;


/* ------------------------------------------------ Module Stuff */

/* Register the initialization status */
static int init_status_flag;
   #define RSHAPER_RX_HOOK_REGISTERED           0x02
   #define RSHAPER_TX_HOOK_REGISTERED           0x04
   #define RSHAPER_NETDEVICE_REGISTERED         0x20


static void __exit rshaper_cleanup(void)
{
    struct rshaper_host *host = NULL;
    unsigned long flags;

    if (init_status_flag&RSHAPER_RX_HOOK_REGISTERED)   /* de-hook rx */
       nf_unregister_hook(&rshaper_rx_hook_ops);

    if (init_status_flag&RSHAPER_TX_HOOK_REGISTERED)   /* de-hook tx */
       nf_unregister_hook(&rshaper_tx_hook_ops);

    spin_lock_irqsave(&rshaper_host_lock, flags);

    while ( (host = first_host) ) { /* free it all */
	first_host = host->next;
	rshaper_kfree_host(host);
	host = NULL;
    }

    spin_unlock_irqrestore(&rshaper_host_lock, flags);

    if (init_status_flag&RSHAPER_NETDEVICE_REGISTERED)
	unregister_netdev(rshaper_dev);
}

static int __init rshaper_init(void)
{
   int result;

    spin_lock_init(&rshaper_host_lock);

    PDEBUG("mode is %i\n", mode);

    /* Register the hook */
    if (mode==1 || mode==2) {
       result = nf_register_hook(&rshaper_rx_hook_ops);
       if (result < 0) {
	   printk(KERN_ERR "rshaper: can't register netfilter hook");
	   rshaper_cleanup();
	   return result;
       }
       init_status_flag |= RSHAPER_RX_HOOK_REGISTERED;
    }

    if (mode==0 || mode==2) {
       result = nf_register_hook(&rshaper_tx_hook_ops);
       if (result < 0) {
	   printk(KERN_ERR "rshaper: can't register netfilter hook");
	   rshaper_cleanup();
	   return result;
       }
       init_status_flag |= RSHAPER_TX_HOOK_REGISTERED;
    }

    rshaper_dev = alloc_netdev(0, "rshaper", rshaper_setup);
    if (!rshaper_dev)
	return -ENOMEM;

    result = register_netdev(rshaper_dev);

    if (result) {
	printk(KERN_ERR "rshaper: can't register netdevice");
	rshaper_cleanup();
	return result;
    }
    init_status_flag |= RSHAPER_NETDEVICE_REGISTERED;

    return 0;
}

module_init(rshaper_init);
module_exit(rshaper_cleanup);

MODULE_LICENSE("GPL");
